//  Copyright 2019 Kakao Corp.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.

import UIKit

enum SelectMode {
    case single
    case multi
}

//MARK: - SelectableItemParamTableViewCell
class SelectableItemParamTableViewCell: ParamTableViewCell {
    var selectMode: SelectMode = .single //default
    var multipleCheckBoxCollectionView: UICollectionView = {
        let layout = UICollectionViewFlowLayout()
        layout.minimumInteritemSpacing = 2
        layout.minimumLineSpacing = 2
        layout.scrollDirection = .horizontal
        let collectionView = UICollectionView(frame: CGRect.zero, collectionViewLayout: layout)
        collectionView.backgroundColor = UIColor.white
        return collectionView
    }()
    
    override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
        super.init(style: style, reuseIdentifier: reuseIdentifier)
        
        multipleCheckBoxCollectionView.delegate = self
        multipleCheckBoxCollectionView.dataSource = self
        multipleCheckBoxCollectionView.showsHorizontalScrollIndicator = false
        multipleCheckBoxCollectionView.register(SelectableCollectionViewCell.self, forCellWithReuseIdentifier: "SelectableCollectionViewCell")
        contentView.addSubview(multipleCheckBoxCollectionView)
        
        addSubviewsConstraint()
    }

    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override func addSubviewsConstraint() {
        super.addSubviewsConstraint()
        multipleCheckBoxCollectionView.translatesAutoresizingMaskIntoConstraints = false
        multipleCheckBoxCollectionView.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        multipleCheckBoxCollectionView.leadingAnchor.constraint(equalTo: paramKeyLabel.trailingAnchor, constant: 4).isActive = true
        multipleCheckBoxCollectionView.trailingAnchor.constraint(equalTo: contentView.trailingAnchor, constant: -4).isActive = true
        multipleCheckBoxCollectionView.heightAnchor.constraint(equalTo: contentView.heightAnchor).isActive = true
    }

    override func layoutSubviews() {
        super.layoutSubviews()
    }

    override func prepareForReuse() {
        super.prepareForReuse()
        
        multipleCheckBoxCollectionView.delegate = nil
        multipleCheckBoxCollectionView.dataSource = nil
    }
    
    func paramItemValues(_ paramItem:ParamItem?) -> [String]? {
        guard let paramItem = paramItem else { return [] }
        
        var apiParamValues : [String]? = nil
        switch paramItem.value {
        case .multiSelectableItem(let values):
            apiParamValues = values
        case .singleSelectableItem(let values):
            apiParamValues = values
        default:
            break
        }
        return apiParamValues
    }
    
    override func setParamItem(_ paramItem:ParamItem) {
        super.setParamItem(paramItem)
        setNeedsLayout()
    }
    
    func selectItem(index:Int) {
        _ = paramItem?.select(index:index)
    }
    
    override func selectItem() {
        super.selectItem()
    }
}

extension SelectableItemParamTableViewCell: UICollectionViewDelegate, UICollectionViewDataSource {
    func collectionView(_ collectionView: UICollectionView, numberOfItemsInSection section: Int) -> Int {
        return paramItemValues(self.paramItem)?.count ?? 0
    }
    
    func collectionView(_ collectionView: UICollectionView, cellForItemAt indexPath: IndexPath) -> UICollectionViewCell {
        guard let paramValues = paramItemValues(self.paramItem) else {
            return SelectableCollectionViewCell()
        }
        
        if let cell = collectionView.dequeueReusableCell(withReuseIdentifier: "SelectableCollectionViewCell", for: indexPath) as? SelectableCollectionViewCell {
            cell.titleLabel.text = paramValues[indexPath.item]
            cell.backgroundColor = .white
            cell.layer.cornerRadius = CommonConstants.paramUIRadius
            cell.layer.masksToBounds = true
            
            if self.paramItem?.selectedIndex.contains(indexPath.item) == true {
                cell.updateCollectionViewCellUI(selected:true)
            }
            return cell
        }
        return SelectableCollectionViewCell()
    }
    
    func collectionView(_ collectionView: UICollectionView, didSelectItemAt indexPath: IndexPath) {
        if let collectionViewCell = collectionView.cellForItem(at: indexPath) as? SelectableCollectionViewCell {
            //model update
            guard let paramItem = self.paramItem else { return }
            let selectedAtIndex = paramItem.select(index: indexPath.item)
            
            //UI update
            updateSwitchUI()
            if selectMode == .single {
                resetCollectionViewCellUI()
            }
            collectionViewCell.updateCollectionViewCellUI(selected:selectedAtIndex)
        }
    }
    
    func resetCollectionViewCellUI() {
        if let visibleCells = self.multipleCheckBoxCollectionView.visibleCells as? [SelectableCollectionViewCell] {
            visibleCells.forEach { cell in
                cell.setSelected(false)
            }
        }
    }
}

extension SelectableItemParamTableViewCell: UICollectionViewDelegateFlowLayout {
    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, insetForSectionAt section: Int) -> UIEdgeInsets {
            return UIEdgeInsets(top: 0, left: CommonConstants.paramCellSpacing, bottom: 0, right: CommonConstants.paramCellSpacing)
        }
    
    func collectionView(_ collectionView: UICollectionView, layout collectionViewLayout: UICollectionViewLayout, sizeForItemAt indexPath: IndexPath) -> CGSize {
        guard let paramValues = paramItemValues(self.paramItem) else { return CGSize(width: 100, height: 40) }
        let valueText : String = paramValues[indexPath.item]
        let width = valueText.widthOfString(usingFont:UIFont.systemFont(ofSize: CommonConstants.paramCollectionViewCellFontSize, weight: .regular))
        return CGSize(width: width + (CommonConstants.paramUIRadius * 2), height: 40)
    }
}

//MARK: - EditableTextParamTableViewCell

class EditableTextParamTableViewCell : ParamTableViewCell, UITextFieldDelegate {
    var paramValueTextField: UITextField = {
        let textField = UITextField(frame: CGRect.zero)
        return textField
    }()
    
    override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
        super.init(style: style, reuseIdentifier: reuseIdentifier)
        paramValueTextField.setLeftPaddingPoints(CommonConstants.paramCellSpacing)
        paramValueTextField.setRightPaddingPoints(CommonConstants.paramCellSpacing)
        paramValueTextField.backgroundColor = UIColor.init(red: 0.1, green: 0.1, blue: 0.1, alpha: 0.1)
        paramValueTextField.layer.cornerRadius = CommonConstants.paramUIRadius
        paramValueTextField.layer.masksToBounds = true
        paramValueTextField.delegate = self
        paramValueTextField.textColor = UIColor.black
        paramValueTextField.adjustsFontSizeToFitWidth = true
        paramValueTextField.attributedPlaceholder = NSAttributedString(
            string: "Placeholder Text",
            attributes: [NSAttributedString.Key.foregroundColor: UIColor.lightGray]
        )
        paramValueTextField.autocapitalizationType = .none
        paramValueTextField.addTarget(self, action: #selector(textFieldDidChange), for: .editingChanged)
        paramValueTextField.clearButtonMode = .whileEditing
        paramValueTextField.rightView = nil
        paramValueTextField.rightViewMode = .never

        if let clearButton = paramValueTextField.value(forKeyPath: "_clearButton") as? UIButton {
            clearButton.setImage(clearButton.image(for: .normal)?.withRenderingMode(.alwaysTemplate), for: .normal)
            clearButton.tintColor = .white
        }
        self.contentView.addSubview(self.paramValueTextField)
        
        addSubviewsConstraint()
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override func addSubviewsConstraint() {
        super.addSubviewsConstraint()
        paramValueTextField.translatesAutoresizingMaskIntoConstraints = false
        paramValueTextField.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        paramValueTextField.leadingAnchor.constraint(equalTo: paramKeyLabel.trailingAnchor, constant: 4).isActive = true
        paramValueTextField.trailingAnchor.constraint(equalTo: contentView.trailingAnchor, constant: -20).isActive = true
        paramValueTextField.heightAnchor.constraint(equalToConstant: CommonConstants.paramValueAreaHeight).isActive = true
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
    }
    
    override func prepareForReuse() {
        super.prepareForReuse()
        paramValueTextField.text = nil
        paramValueTextField.delegate = nil
        paramValueTextField.isEnabled = true
        paramValueTextField.backgroundColor = UIColor.init(red: 0.1, green: 0.1, blue: 0.1, alpha: 0.1)
    }
    
    override func setParamItem(_ paramItem:ParamItem) {
        super.setParamItem(paramItem)
        
        if case .editableText(let defaultValue) = paramItem.value {
            paramValueTextField.text = defaultValue
            paramValueTextField.placeholder = paramItem.placeHolderValue
            paramValueTextField.isEnabled = paramItem.editable
            if (paramItem.editable == false) {
                paramValueTextField.backgroundColor = UIColor.white
            }
        }
    }
    
    override func selectItem() {
        super.selectItem()
    }
    
    func updateItem(text:String) {
        paramItem?.update(newValue: text)
        //self.selectItem() //값 바꾸는것과 선택은 별개
    }
    
    // UITextField Delegates
    func textFieldDidBeginEditing(_ textField: UITextField) {
    }
    
    func textFieldDidEndEditing(_ textField: UITextField) {
        if let text = textField.text {
            updateItem(text:text)
            
            if (text.count == 0) {
                paramItem?.select(false)
                updateSwitchUI()
            }
        }
    }
    
    func textFieldShouldBeginEditing(_ textField: UITextField) -> Bool {
        return true;
    }
    
    func textFieldShouldClear(_ textField: UITextField) -> Bool {
        return true;
    }
    
    func textFieldShouldEndEditing(_ textField: UITextField) -> Bool {
        return true;
    }
    
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange, replacementString string: String) -> Bool {
        return true;
    }
    
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        textField.resignFirstResponder();
        return true;
    }
    
    @objc func textFieldDidChange() {
        if let text = paramValueTextField.text {
            updateItem(text:text)
        }
        paramItem?.select(false)
        updateSwitchUI()
    }
    
}

//MARK: - ParamCell
class ParamTableViewCell: UITableViewCell {

    var paramItem: ParamItem?
    
    var useParamSwitch: UISwitch = {
        let paramSwitch = UISwitch()
        return paramSwitch
    }()
    
    var paramKeyLabel: UILabel = {
        let label = UILabel(frame: CGRect.zero)
        return label
    }()
    
    override init(style: UITableViewCell.CellStyle, reuseIdentifier: String?) {
        super.init(style: style, reuseIdentifier: reuseIdentifier)

        self.useParamSwitch.isOn = false
        self.useParamSwitch.addTarget(self, action: #selector(onValueChanged(_:)), for: .valueChanged)
        self.contentView.addSubview(self.useParamSwitch)
        
        self.paramKeyLabel.backgroundColor = UIColor.init(white: 1.0, alpha: 0)
        self.paramKeyLabel.font = UIFont.boldSystemFont(ofSize: 18)
        self.paramKeyLabel.textColor = UIColor.black
        self.paramKeyLabel.textAlignment = .right
        self.paramKeyLabel.adjustsFontSizeToFitWidth = true
        self.contentView.addSubview(self.paramKeyLabel)
        
        self.accessoryType = .none
        self.contentView.backgroundColor = UIColor.white
        
        self.selectionStyle = .none
    }
    
    func addSubviewsConstraint() {
        useParamSwitch.translatesAutoresizingMaskIntoConstraints = false
        useParamSwitch.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        useParamSwitch.leadingAnchor.constraint(equalTo: contentView.leadingAnchor, constant: CommonConstants.paramCellSpacing).isActive = true
        
        paramKeyLabel.translatesAutoresizingMaskIntoConstraints = false
        paramKeyLabel.centerYAnchor.constraint(equalTo: contentView.centerYAnchor).isActive = true
        paramKeyLabel.leadingAnchor.constraint(equalTo: useParamSwitch.trailingAnchor, constant: CommonConstants.paramCellSpacing).isActive = true
        paramKeyLabel.widthAnchor.constraint(equalToConstant: CommonConstants.paramKeyAreaWidth).isActive = true
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("Never will happen")
    }
    
    override func layoutSubviews() {
        super.layoutSubviews()
    }
    
    override func prepareForReuse() {
        super.prepareForReuse()
        
        self.useParamSwitch.isOn = false
        self.paramKeyLabel.text = nil
        self.accessoryType = .none
    }
    
    func setParamItem(_ paramItem:ParamItem) {
        self.paramItem = paramItem
        self.paramKeyLabel.text = paramItem.key
    }
    
    func selectItem() {
        paramItem?.select()
    }
    
    @objc func onValueChanged(_ switch: UISwitch) {
        switch self.paramItem?.value {
        case .editableText(_):
            self.selectItem()
            self.updateSwitchUI()
        case .multiSelectableItem(_):
            fallthrough
        case .singleSelectableItem(_):
            if self.paramItem?.isSelected() == true {
                self.paramItem?.resetSelect()
                (self as? SelectableItemParamTableViewCell)?.resetCollectionViewCellUI()
            }
            self.updateSwitchUI()
            break
        default:
            break
        }
    }
    
    func updateSwitchUI() {
        if let paramItem = self.paramItem {
            useParamSwitch.setOn(paramItem.selected, animated: true)
        }
    }
}
